#!BPY

# """
# Name: 'Edge tune'
# Blender: 241
# Group: 'Mesh'
# Tooltip: 'Freehand vertslide'
# """
#
#
# version 2.0
# march 9, 2007
# by Gert De Roost		http://users.telenet.be/EWOCprojects
# Under GPL license.

__author__ = 'Gert De Roost'
__version__ = '2.0 march 9, 2007'
__url__ = ["EWOCprojects, http://users.telenet.be/EWOCprojects"]
__email__ = ["Gert De Roost, ewocprojects@yahoo.com"]
__bpydoc__ = """\
This script is an implementation of the concept of sliding vertices around on edges. 
It is used to finetune/redraw edges/edgeloops through the process of sliding vertices. 
It can be used to slide anything anywhere.
To my knowledge this is a new concept in 3D modelling. Try it and you will see how it can impact your modelling habits. 
You are able to tune vertices by sliding by freehand-redrawing them on the edges they are part of.


This is version 2 of the script, implementing a new graphical user interface and having gone to a complete rewrite of the code,
lifting almost all shortcomings of the old code.


Documentation
The script will work on any vertice/edge/face-selection. 
Make a selection, invoke script.  The selection will be visualized in yellow.
Press and hold left-mouse button and draw freely across the in red visualized, so called "slide-edges".
The respective selected vertices will change position on the slide-edge to the new position you choose
by moving over it with the left mouse-button pressed.
Press ENTER/RETURN to finalize the operation.
Just press the right-mouse-button to cancel the script operation.

HINT: EdgeTune is also vertex-slide.

Version 2 now works in both orthogonal and perspective mode.

Version 2 now doesn't mess with the undo list anymore, plus the script operation can be
undone by pressing standard Ctrl-Z.

Do BEWARE still: EdgeTune, like every other script that changes the mesh, will decompose all FGons in the mesh!
"""


from math import sqrt
import Blender
from Blender.BGL import *
from Blender.Mathutils import *
from Blender import Scene, Draw, Window, Object, NMesh, Text
from os import sep




def optimaloff():
	global optimalswitch, meshmodes
	# Turn off Optimal mode if set.
	optimalswitch = 0
	meshmodes = []
	modeflags = mesh.getMode()
	if modeflags & 256:
		optimalswitch = 1
		if modeflags & 2:
			meshmodes.append("NoVNormalsFlip")
		if modeflags & 4:
			meshmodes.append("TwoSided")
		if modeflags & 32:
			meshmodes.append("AutoSmooth")
		if modeflags & 128:
			meshmodes.append("SubSurf")
		mesh.setMode(*meshmodes)
		# Update display.
		mesh.update()
		Window.EditMode(1)
		Window.Redraw(Window.Types.VIEW3D)
		Window.EditMode(0)
				

def optimalonmaybe():
	global optimalswitch, meshmodes
	# Turn Optimal mode back on if necessary.
	if optimalswitch:
		meshmodes.append("Optimal")
		mesh.setMode(*meshmodes)
		# Update display.
		mesh.update()
		Window.EditMode(1)
		Window.Redraw(Window.Types.VIEW3D)
		Window.EditMode(0)


def getcurrentarea():
	while 1:
		mousecoords = Window.GetMouseCoords()
		for area in Window.GetScreenInfo(Window.Types.VIEW3D):
			x1, y1, x2, y2 = area.get('vertices')
			if (x1 < mousecoords[0] < x2) and (y1 < mousecoords[1] < y2):
				return area

def getscreencoords(vector):
	vector = Vector(vector + [1])
	pvector = vector * Blender.matrix
	pvector[0] /= pvector[3]
	pvector[1] /= pvector[3]
	return [Blender.viewport[0] + (pvector[0] * Blender.viewport[2]), Blender.viewport[1] + (pvector[1] * Blender.viewport[3]), pvector[2]]


def loadedgetunedisplay():
	global displayscript
	failed = 0
	if Blender.Get("scriptsdir") != None:
		try:
			displayscript = Text.Load("".join([Blender.Get("scriptsdir"), sep+"edgetunedisplay.py"]))
		except:
			pass
			failed = 1
	else: failed = 1 
			
	if Blender.Get("uscriptsdir") != None:
		try:
			displayscript = Text.Load("".join([Blender.Get("uscriptsdir"), sep+"edgetunedisplay.py"]))
		except:
			pass
			failed == 2
	elif failed == 1:
		failed = 2
	if failed == 2:
		Draw.PupMenu("Help-script 'edgetunedisplay.py' not installed correctly.")
		
	scene.addScriptLink("edgetunedisplay.py","Redraw")


def unloadedgetunedisplay():
	global displayscript
	# Remove the "edgetunedisplay" scriptlink and text from the scene.
	
	# Store currently running scriptlinks.
	framechangedlist = scene.getScriptLinks("FrameChanged")
	onloadlist = scene.getScriptLinks("OnLoad")
	onsavelist = scene.getScriptLinks("OnSave")
	redrawlist = scene.getScriptLinks("Redraw")
	try:
		redrawlist.pop(redrawlist.index("edgetunedisplay.py"))
	except:
		pass
	# Remove all links from scene.
	scene.clearScriptLinks()
	# Restore saved scriptlinks.
	if framechangedlist != None:
		for script in framechangedlist:
			scene.addScriptLink(script, "FrameChanged")
	if onloadlist != None:
		for script in onloadlist:
			scene.addScriptLink(script, "OnLoad")
	if onsavelist != None:
		for script in onsavelist:
			scene.addScriptLink(script, "OnSave")
	if redrawlist != None:
		for script in redrawlist:
			scene.addScriptLink(script, "Redraw")
	Text.unlink(displayscript)


def mul(a, b):
	# a is vector, b is matrix
	r = [0, 0, 0]
	r[0] = a[0]*b[0][0]+a[1]*b[1][0]+a[2]*b[2][0]+b[3][0]
	r[1] = a[0]*b[0][1]+a[1]*b[1][1]+a[2]*b[2][1]+b[3][1]
	r[2] = a[0]*b[0][2]+a[1]*b[1][2]+a[2]*b[2][2]+b[3][2]
	return r
	








#################
#				#
# Main program. #
#				#
#################


while Window.GetMouseButtons() :      
	pass

# Do this to refresh mesh if just subdivided.
Window.EditMode(0)
Window.EditMode(1)


# Get mesh.
scene = Scene.GetCurrent()
obj = Object.GetSelected()
obj = obj[0]
Blender.selobj = obj

mesh = obj.getData()


# Wait for VIEW3D to be selected if multiple views.
workspaces = Window.GetScreenInfo(Window.Types.VIEW3D)
if len(workspaces) == 0:
	Draw.PupMenu("No 3D views, press right mouse button to quit.")
elif len(workspaces) == 1:
	area = workspaces[0]
else:
	while not(Window.GetMouseButtons()):
		pass
	while Window.GetMouseButtons():
		pass
	area = getcurrentarea()

# Calculate viewport, get transformation matrix.
areaid = area.get("id")
Blender.xmin, Blender.ymin, Blender.xmax, Blender.ymax = area.get("vertices")
Blender.viewport = [(Blender.xmax + Blender.xmin) / 2, (Blender.ymax + Blender.ymin) / 2, (Blender.xmax - Blender.xmin) / 2, (Blender.ymax - Blender.ymin) / 2]
Window.QHandle(areaid)
Blender.matrix = obj.getMatrix() * Window.GetPerspMatrix()



# seledges: selected edges list
# edgestofacesd: dictionary with edge key and faces list that contain edge as value 
# Blender.selverts: selected verts list per edge
# Blender.selcoords: selected verts coordinate list per edge
seledges = []
edgestofacesd = {}
Blender.selverts = []
Blender.selcoords = []
for edge in mesh.edges:
	if edge.v1.sel and edge.v2.sel:
		seledges.append(edge)
		edgestofacesd[edge] = []
		Blender.selverts.append([edge.v1, edge.v2])
		x1, y1, dummy = getscreencoords(edge.v1.co[:])
		x2, y2, dummy = getscreencoords(edge.v2.co[:])
		Blender.selcoords.append([[x1, y1],[x2, y2]])
		for face in mesh.faces:
			if (edge.v1 in face.v) and (edge.v2 in face.v):
				edgestofacesd[edge].append(face)


# selverts: selected verts list
# slideedges: slideedges list
# Blender.slideverts: slideverts list per edge
# Blender.slidecoords: slideverts coordinate list per edge
Blender.vertd = {}
Blender.slideverts = []
Blender.slidecoords = []
slideedges = []
selverts = []
for vert in mesh.verts:
	if vert.sel:
		Blender.vertd[vert] = vert.co[:]
		selverts.append(vert)
		for edge in mesh.edges:
			if vert == edge.v1 or vert == edge.v2:
				if not(edge in seledges):
					slideedges.append(edge)
					Blender.slideverts.append([edge.v1, edge.v2])
					x1, y1, dummy = getscreencoords(edge.v1.co[:])
					x2, y2, dummy = getscreencoords(edge.v2.co[:])
					Blender.slidecoords.append([[x1, y1], [x2, y2]])

# Box out single vertices.
Blender.singles = []
Blender.boxes = []
for vert in selverts:
	single = 1
	for edge in seledges:
		if vert == edge.v1 or vert == edge.v2:
			single = 0
			break
	if single == 1:
		print "single"
		Blender.singles.append(vert)
		Blender.boxes.append(getscreencoords(vert.co[:])[:2])

# Quit if no selected verts.
if len(selverts) == 0:
	Draw.PupMenu("Nothing selected, press right mouse button to quit.")


# Start display help-script.
optimaloff()
Blender.viewchange = 0
Blender.mouseover = 0	
Blender.x1, Blender.x2, Blender.y1, Blender.y2 = -1, -1, -1, -1
loadedgetunedisplay()




mx = my = -1
vertd = {}



while Window.GetMouseButtons() != 4:		# right mouse button cancels

	Blender.viewchange = 0

	# Main event loop.
	evt, val = Window.QRead()

	if evt == Draw.REDRAW: Blender.Redraw(-1)
	elif evt in [Draw.LEFTSHIFTKEY, Draw.MIDDLEMOUSE, Draw.WHEELDOWNMOUSE, Draw.WHEELUPMOUSE]:
		# Rotating / panning / zooming 3D view is handled here.
		Blender.viewchange = 1
		Window.QAdd(areaid, evt, val)
		Window.QHandle(areaid)
		Window.QRedrawAll()
		continue
	elif evt == Draw.RETKEY:
		# Break from while loop if ENTER pressed.
		# Do slide operation in mesh.
		for vert in Blender.vertd.keys():
			vert.co[0] = Blender.vertd[vert][0]
			vert.co[1] = Blender.vertd[vert][1]
			vert.co[2] = Blender.vertd[vert][2]
		# Update display.
		Window.EditMode(0)
		mesh.update()
		Window.EditMode(1)					
		Window.Redraw(Window.Types.VIEW3D)
		break
	elif evt in [Draw.MOUSEX, Draw.MOUSEY]:
		oldmx = mx
		oldmy = my
		mx, my = Window.GetMouseCoords()
		hoveredge = None

		# First check mouse is in bounding box edge of which edges.
		testscrl = []
		for edge in slideedges:
			x1, y1, dummy = getscreencoords(edge.v1.co[:])
			x2, y2, dummy = getscreencoords(edge.v2.co[:])
			if x1 < x2:
				lwpx = x1 - 5
				uppx = x2 + 5
			else:
				lwpx = x2 - 5
				uppx = x1 + 5
			if y1 < y2:
				lwpy = y1 - 5
				uppy = y2 + 5
			else:
				lwpy = y2 - 5
				uppy = y1 + 5		
			if (((x1 < mx < x2) or (x2 < mx < x1)) and (lwpy < my < uppy)) or (((y1 < my < y2) or (y2 < my < y1)) and (lwpx < mx < uppx)):
				testscrl.append(edge)

		# Then check these edges to see if mouse is on one of them.
		allhoveredges = []
		hovering = 0
		zmin = 1e10
		if testscrl != []:
			for edge in testscrl:
				x1, y1, z1 = getscreencoords(edge.v1.co[:])
				x2, y2, z2 = getscreencoords(edge.v2.co[:])

				if x1 == x2 and y1 == y2:
					dist = sqrt((mx - x1)**2 + (my - y1)**2)
				else:
					dist = ((mx - x1)*(y2 - y1) - (my - y1)*(x2 - x1)) / sqrt((x2 - x1)**2 + (y2 - y1)**2)

				if -5 < dist < 5:
					allhoveredges.append(edge)
					if hoveredge != None and ((z1 + z2) / 2) > zmin:
						pass
					else:
						hovering = 1
						hoveredge = edge
						zmin = (z1 + z2) / 2
						Blender.mouseover = 1
						x1, y1, dummy = Blender.x1, Blender.y1, dummy = getscreencoords(hoveredge.v1.co[:])
						x2, y2, dummy = Blender.x2, Blender.y2, dummy = getscreencoords(hoveredge.v2.co[:])
		if hovering == 0:
			Blender.mouseover = 0
			Blender.x1, Blender.x2, Blender.y1, Blender.y2 = -1, -1, -1, -1
		if oldmx != mx or oldmy != my:
			Blender.Redraw()
		




		mbns = Window.GetMouseButtons()
		
		if hoveredge != None and mbns == 1:
			# Find projection mouse perpend on edge.
			if x1 == x2:	x1 += 1e-6
			if y1 == y2:	y1 += 1e-6
			a = (x2 - x1) / (y2 - y1)
			x = ((x1 / a) + (mx * a) + my - y1) / ((1 / a) + a)
			y = ((mx - x) * a) + my
			# Calculate relative position on edge and adapt screencoords accoringly.
			div = (x - x1) / (x2 - x1)
			if hoveredge.v1 in selverts:
				vert = hoveredge.v1
				vert2 = hoveredge.v2
			else:
				vert = hoveredge.v2
				vert2 = hoveredge.v1
			x1, y1, dummy = getscreencoords(hoveredge.v1.co[:])
			x2, y2, dummy = getscreencoords(hoveredge.v2.co[:])
			coords = [((x2 - x1) * div ) + x1, ((y2 - y1) * div ) + y1]
			for verts in Blender.selverts:
				if vert == verts[0]:
					Blender.selcoords[Blender.selverts.index(verts)][0] = coords
				elif vert == verts[1]:	
					Blender.selcoords[Blender.selverts.index(verts)][1] = coords
			if vert in Blender.singles:
				Blender.boxes[Blender.singles.index(vert)] = coords
			# Calculate and remember new vert 3D coordinates.  Set in physical mesh at end of program.		
			vx1, vy1, vz1 = hoveredge.v1.co[:]
			vx2, vy2, vz2 = hoveredge.v2.co[:]
			Blender.vertd[vert] = [((vx2 - vx1) * div ) + vx1, ((vy2 - vy1) * div ) + vy1, ((vz2 - vz1) * div ) + vz1]
			Window.Redraw(Window.Types.VIEW3D)
		 	
		continue



optimalonmaybe()
unloadedgetunedisplay()
