# version 2
# March 9, 2007
# by Gert De Roost		http://users.telenet.be/EWOCprojects2
# --------------------------------------------------------------------------
# ***** BEGIN GPL LICENSE BLOCK *****
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# If you have Internet access, you can find the license text at
# http://www.gnu.org/licenses/gpl.txt,
# if not, write to the Free Software Foundation,
# Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# ***** END GPL LICENCE BLOCK *****
# --------------------------------------------------------------------------


import Blender
from Blender.Mathutils import *
from Blender.BGL import *
from Blender import Window


def getscreencoords(vector):
	vector = Vector(vector + [1])
	pvector = vector * Blender.matrix
	pvector[0] /= pvector[3]
	pvector[1] /= pvector[3]
	return [Blender.viewport[0] + (pvector[0] * Blender.viewport[2]), Blender.viewport[1] + (pvector[1] * Blender.viewport[3]), pvector[2]]


def adaptcoordlists():
	# Transform lists of coords to draw.
	Blender.matrix = Blender.selobj.getMatrix() * Window.GetPerspMatrix()
	for posn in range(len(Blender.selverts)):
		Blender.selcoords[posn] = [getscreencoords(Blender.vertd[Blender.selverts[posn][0]][:])[:2], getscreencoords(Blender.vertd[Blender.selverts[posn][1]][:])[:2]]
	for posn in range(len(Blender.slideverts)):
		Blender.slidecoords[posn] = [getscreencoords(Blender.slideverts[posn][0].co[:])[:2],  getscreencoords(Blender.slideverts[posn][1].co[:])[:2]]
	for posn in range(len(Blender.singles)):
		Blender.boxes[posn] = getscreencoords(Blender.vertd[Blender.singles[posn]][:])[:2]
		

def saveOGLsetup():
	# Backup OpenGL settings.

	glMatrixMode(GL_MODELVIEW)
	glPushMatrix()
	glMatrixMode(GL_PROJECTION)
	glPushMatrix()


def restoreOGLsetup():
	# Restore initial OpenGL settings.

	glMatrixMode(GL_PROJECTION)
	glPopMatrix()
	glMatrixMode(GL_MODELVIEW)
	glPopMatrix()


def initOGL():
	glMatrixMode(GL_PROJECTION)
	glLoadIdentity()
	glOrtho(Blender.xmin, Blender.xmax, Blender.ymin, Blender.ymax, -1, 2)
	glMatrixMode(GL_MODELVIEW)
	glLoadIdentity()
	glScissor(Blender.xmin, Blender.ymin, Blender.xmax - Blender.xmin, Blender.ymax - Blender.ymin)




saveOGLsetup()
initOGL()

if Blender.viewchange:
	adaptcoordlists()

if Blender.slideverts != []:
	# Draw single verts as boxes.
	glColor3f(1.0,1.0,0)
	for vertcoords in Blender.boxes:
		glBegin(GL_POLYGON)
		x, y = vertcoords
		glVertex2f(x-2, y-2)
		glVertex2f(x-2, y+2)
		glVertex2f(x+2, y+2)
		glVertex2f(x+2, y-2)
		glEnd()

	# Accentuate selected edges.
	glColor3f(1.0,1.0,0)
	for posn in range(len(Blender.selcoords)):
		glBegin(GL_LINES)
		x, y = Blender.selcoords[posn][0]
		glVertex2f(x, y)
		x, y = Blender.selcoords[posn][1]
		glVertex2f(x, y)
		glEnd()

	# Draw slide-edges.
	glColor3f(1.0,0,0)
	for posn in range(len(Blender.slidecoords)):
		glBegin(GL_LINES)
		x, y = Blender.slidecoords[posn][0]
		glVertex2f(x, y)
		x, y = Blender.slidecoords[posn][1]
		glVertex2f(x, y)
		glEnd()

# Draw mouseover highlighting.
if Blender.mouseover:
	glColor3f(0,0,1.0)
	glBegin(GL_LINES)
	x,y = Blender.x1, Blender.y1
	glVertex2f(x,y)
	x,y = Blender.x2, Blender.y2
	glVertex2f(x,y)
	glEnd()

restoreOGLsetup()
